from .view import View


class ErrorView(View):
    def __init__(self):
        super().__init__()
        self.title("Error")
        self.errorMessage = "Error"

    def errorMessage(self):
        return self.errorMessage

    def set_errorMessage(self, error_msg):
        self.errorMessage = error_msg
        return self

    def asDictionaryForExport(self):
        exportData = super().asDictionaryForExport()
        exportData["viewName"] = "GtPhlowViewErrorViewSpecification"
        exportData["dataTransport"] = 1
        exportData["errorMessage"] = self.errorMessage
        return exportData



class View:
	def __init__(self):
		self.viewTitle = "Unknown"
		self.viewPriority = 1
		self.accessor = None

	def title(self, title):
		self.viewTitle = title
		return self
	
	def priority(self, priority):
		self.viewPriority = priority
		return self
	
	def getTitle(self):
		return self.viewTitle
	
	def set_accessor(self, accessor_function):
		self.accessor = accessor_function
		return self
	
	def asDictionaryForExport(self):
		return {
			"title": self.viewTitle,
			"priority": self.viewPriority
		}



import msgpack
import socket
import _thread
import threading
import time
import sys

from .stoppable_thread import StoppableThread
from .msgpack_serializer import MsgPackSerializer
from .msgpack_serializer import addMapping as msgpack_serializer_addMapping
from .bridge_utils import random_str
from .bridge_utils import log_msg

# Messages supported by this sockets must be Dictionaries. This is because we use special key __sync to know if it is 
# a synchronized message or not. If it is we hook a semaphore to that id under the __sync key and after we receive the 
# value we store there the return message and signal the semaphore.
class MsgPackSocketPlatform:

    def __init__(self, port):
        self.port = port
        self.server = None
        self.connection = None
        self.serializer = MsgPackSerializer()
        self.unpacker = msgpack.Unpacker(raw=False)
        self.packer = msgpack.Packer(use_bin_type=True)
        self.sync_table = {}
        self.async_handlers = {}
    
    def addMapping(self, key_type, mapping_function):
        msgpack_serializer_addMapping(key_type, mapping_function)

    def set_handler(self, msg_type, async_handler):
        self.async_handlers[msg_type] = async_handler

    def prim_handle(self):
        try:
            log_msg("HANDLER (MsgPackSocketPlatform): loop func")
            data = self.getConnection().recv(2048)
            if len(data) == 0:
                log_msg("HANDLER (MsgPackSocketPlatform): received zero bytes, done")
                self.connection.close()
                self.connection = None
            else:
                self.unpacker.feed(data)
                for msg in self.unpacker:
                    log_msg("HANDLER (MsgPackSocketPlatform): prim handle message")
                    self.prim_handle_msg(msg)
        except OSError:
            log_msg("HANDLER (MsgPackSocketPlatform): OSError " + str(err))
            self.stop()
            sys.exit()
            exit(-1)
        except Exception as err:
            log_msg("HANDLER (MsgPackSocketPlatform): ERROR " + str(err))

    def setup_func(self):
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.bind(('localhost', self.port))
        self.server.listen()
        
    def getConnection(self):
        if (self.connection == None):
            self.connection, addr = self.server.accept()
        return self.connection

    def stop(self):
        if self.thread is not None:
            self.thread.stop()
        if self.server is not None:
            self.server.close()
            self.server = None

    def send_answer(self, msg, answer):
        if answer['type'] != msg['type']:
            raise Exception('Type mismatch')
        answer['__sync'] = msg['__sync']
        self.send_async_message(answer)
    
    def is_running(self):
        return self.server != None
    
    def prim_handle_msg(self, raw_msg):
        msg = raw_msg
        msg_type = msg['type'] 
        if msg_type in self.async_handlers:
            self.async_handlers[msg['type']](msg)
        elif is_sync_msg(msg):
            sync_id = message_sync_id(msg)
            semaphore = self.sync_table[sync_id]
            self.sync_table[sync_id] = msg
            semaphore.release()
        else:
            log_msg("Error! Msg couldnt be handled")
            raise Exception('Message couldn''t be handled')
    
    def start(self):
        self.thread = StoppableThread(
            loop_func= self.prim_handle,
            setup_func= self.setup_func)
        self.thread.start()
        # time.sleep(.1)

    def send_async_message(self, msg):
        self.getConnection().send(self.packer.pack(msg))
    
    def send_sync_message(self, msg):
        sync_id = mark_message_as_sync(msg)
        semaphore = threading.Semaphore(value=0)
        self.sync_table[sync_id] = semaphore
        self.send_async_message(msg)
        semaphore.acquire()
        ans = self.sync_table[sync_id]
        del self.sync_table[sync_id]
        return ans

def is_sync_msg(msg):
    return '__sync' in msg
    
def message_sync_id(msg):
    return msg['__sync']
    
def mark_message_as_sync(msg):
    sync_id = random_str()
    msg['__sync'] = sync_id
    return sync_id

def build_service(port, pharo_port, feed_callback):
    service = MsgPackSocketPlatform(port)
    service.set_handler('ENQUEUE',feed_callback)
    service.set_handler('IS_ALIVE', lambda msg: service.send_answer(msg, {'type': 'IS_ALIVE'}))
    return service



from flask import Flask, request
import http.client
import json
import threading
import sys
import logging
import requests
from .json_encoder import JsonSerializer
from .bridge_utils import random_str
from .bridge_utils import log_msg

class FlaskMsgService:

    def __init__(self, port, pharo_port, feed_callback):
        self.serializer = json_encoder.JsonSerializer()
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)
        self.thread = None
        self.port = port
        self.pharo_port = pharo_port
        self.feed_callback = feed_callback
        self.app = Flask('PythonBridge')
        self.app.use_reloader=False
        self.session = requests.Session()
        self.session.trust_env = True

        @self.app.route("/ENQUEUE", methods=["POST"])
        def eval_expression():
            data = request.get_json(force=True)
            self.feed_callback(data)
            return "{}"

        @self.app.route("/IS_ALIVE", methods=["POST"])
        def status_endpoint():
            return "{}"
        
    def addMapping(self, key_type, mapping_function):
        json_encoder.addMapping(key_type, mapping_function)

    def _start(self):
        try:
            self.app.run(port=self.port)
        except OSError as err:
            log_msg('Critical Error:' + str(err))
            exit(42)
    
    def start(self):
        self.thread = threading.Thread(target=self._start, args=())
        self.thread.daemon = True
        self.thread.start()

    def is_running(self):
        return self.thread != None

    def stop(self):
        pass

    def send_async_message(self, msg):
        self.send_sync_message(msg)
    
    def send_sync_message(self, msg):
        msg['__sync'] = random_str()
        log_msg("SYNC_MSG: " + json.dumps(msg))
        response = self.session.post(
            'http://localhost:' + str(self.pharo_port) + '/' + msg['type'], 
            data=json.dumps(msg),
            headers={'content-type': 'application/json'}, 
            allow_redirects=True).content.decode('utf-8')
        log_msg("SYNC_ANS: " + response)
        return json.loads(response)

def build_service(port, pharo_port, feed_callback):
    return FlaskMsgService(port, pharo_port, feed_callback)



from .view_list import ListView
from .view_columnedlist import ColumnedListView
from .view_columnedtree import ColumnedTreeView
from .view_texteditor import TextEditorView
from .view_forward import ForwardView
from .view_empty import EmptyView

class ViewBuilder:
	def textEditor(self):
		return TextEditorView()

	def list(self):
		return ListView()

	def columnedList(self):
		return ColumnedListView()

	def columnedTree(self):
		return ColumnedTreeView()

	def forward(self):
		return ForwardView()

	def empty(self):
		return EmptyView()


import argparse
import inspect
import threading
import sys

from .bridge_globals import bridge_globals
from .object_registry import registry
from .bridge_utils import log_msg
from .bridge_hooks import notify, notify_observer, notify_error
from .bridge_hooks import serialize, deserialize, observer


def pbbreak():
    print("Breaking now")
    breakpoint()
    print("Continuing")


class EvalCommand:
    statements = ""
    bindings = {}
    commandId = 0

    def __init__(self, commandId, statements, bindings):
        self.statements = statements
        self.commandId = commandId
        self.bindings = bindings

    def execute_using_env(self, env):
        try:
            env.update(self.bindings)
            exec(self.statements, env)
        except Exception as err:
            self.perform_proceed_action(notify_error(err, self))

    def perform_proceed_action(self, actionDict):
        actionSymbol = actionDict['action']
        if actionSymbol == "IGNORE":
            pass
        if actionSymbol == "DROP_QUEUE":
            bridge_globals()['cmd_list'].drop_queue()
        if actionSymbol == "REPLACE_COMMAND":
            commandDict = actionDict["command"]
            bridge_globals()['cmd_list'].push_command_at_first(EvalCommand(
                commandDict["commandId"],
                commandDict["statements"],
                commandDict["bindings"]))

    def command_id(self):
        return self.commandId


def log_stderr_flush(msg):
    print(str(msg), file=sys.stderr, flush=True)


class Logger:
    def log(self, msg):
        log_stderr_flush(msg)


class NoLogger:
    def log(self, msg):
        pass


# This List is thought to be multi-producer and single-consumer. For optimal results wait for push_command return
# value to push another command that depends on the previous one.
class PythonCommandList:
    currentCommandIndex = 0
    commandList = []
    listLock = threading.Lock()
    consumeSemaphore = threading.Semaphore(value=0)

    # This method locks the thread until the command has been succesfully appended to the list. Even though that it
    # has a lock inside, we do not expect long waiting time.
    def push_command(self, aCommand):
        self.listLock.acquire()
        self.commandList.append(aCommand)
        commandIndex = len(self.commandList) - 1
        self.listLock.release()
        self.consumeSemaphore.release()
        return commandIndex

    def push_command_at_first(self, aCommand):
        self.listLock.acquire()
        self.commandList.insert(self.currentCommandIndex, aCommand)
        self.listLock.release()
        self.consumeSemaphore.release()
        return self.currentCommandIndex

    def drop_queue(self):
        self.listLock.acquire()
        self.consumeSemaphore = threading.Semaphore(value=0)
        self.currentCommandIndex = len(self.commandList)
        self.listLock.release()

    def consume_command(self):
        repeatMonitorFlag = True
        while repeatMonitorFlag:
            self.consumeSemaphore.acquire()
            self.listLock.acquire()
            repeatMonitorFlag = False
            if self.currentCommandIndex >= len(self.commandList):
                repeatMonitorFlag = True
                self.listLock.release()
        command = self.commandList[self.currentCommandIndex]
        self.currentCommandIndex += 1
        self.listLock.release()
        return command

    def get_current_command(self):
        if self.currentCommandIndex == 0:
            return None
        self.listLock.acquire()
        command = self.commandList[self.currentCommandIndex - 1]
        self.listLock.release()
        return command

    def get_command_list(self):
        self.listLock.acquire()
        listCopy = self.commandList.copy()
        self.listLock.release()
        return listCopy


def deserialize(text):
    result = bridge_globals()['msg_service'].serializer.deserialize(text)
    log_msg("DESERIALISE (bridge): " + str(result))
    if registry().isProxy(result):
        result = registry().resolve(result['__pyid__'])
    return result


def enqueue_command(data):
    bridge_globals()['cmd_list'].push_command(EvalCommand(
        data["commandId"],
        data["statements"],
        {k: deserialize(v) for k, v in data["bindings"].items()}))


def parse_bridge_cmd_line_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("-p", "--port", required=False,
                    help="port to be used for receiving instructions")
    ap.add_argument("-o", "--pharo", required=True,
                    help="port to be used for sending notifications back to pharo")
    ap.add_argument("-m", "--method", required=False,
                    help="identifier for communication protocol strategy http or msgpack")
    ap.add_argument("--log", required=False, const=True, nargs="?",
                    help="enable logging")
    return vars(ap.parse_args())


def bridge_args():
    return {'port': 9099, 'pharo': 0, 'method': 'msgpack', 'log': True}


def setup_bridge(args):
    bridge_globals()['pharo_port'] = args["pharo"]

    if args["log"]:
        bridge_globals()['logger'] = Logger()
    else:
        bridge_globals()['logger'] = NoLogger()

    bridge_globals()['python_port'] = args["port"]

    msg_service = None
    if args["port"] is None:
        args["port"] = '0'
    if args["method"] is None:
        args["method"] = 'http'
    if args["method"] == 'http':
        from .flask_platform import build_service
        msg_service = build_service(int(args["port"]), int(args["pharo"]), enqueue_command)
    elif args["method"] == 'msgpack':
        from .msgpack_socket_platform import build_service
        msg_service = build_service(int(args["port"]), int(args["pharo"]), enqueue_command)
    else:
        raise Exception("Invalid communication strategy.")
    bridge_globals()['msg_service'] = msg_service


def run_bridge():
    bridge_globals()['msg_service'].start()

    log_stderr_flush("PythonBridge ready")

    bridge_globals()['cmd_list'] = PythonCommandList()
    globalCommandList = bridge_globals()['cmd_list']
    env = dict(globals())

    log_msg("PYTHON: Start consuming commands")
    while True:
        command = globalCommandList.consume_command()
        log_msg("PYTHON: Executing command " + command.command_id())
        log_msg("PYTHON: bindings: " + str(command.bindings))
        log_msg("PYTHON: " + command.statements)
        command.execute_using_env(env)
        log_msg("PYTHON: Finished command execution")


def run_bridge_default():
    log_stderr_flush("PythonBridge starting")

    args = bridge_args()

    log_stderr_flush(args)

    setup_bridge(args)

    run_bridge()


def run_bridge_main():
    log_stderr_flush("PythonBridge starting")

    args = parse_bridge_cmd_line_args()

    log_stderr_flush(args)

    setup_bridge(args)

    run_bridge()


if __name__ == "__main__":
    run_bridge_main()



from .phlow.view_builder import ViewBuilder
from .phlow.view_error import ErrorView


def gtView(func):
    setattr(func, "gtView", True)
    return func


class GtViewedObject:
    def __init__(self, obj):
        self.object = obj

    def getObject(self):
        return self.object

    def getGtViewMethodNames(self):
        # those defined by me
        allMyAttributes = dir(self)
        allMyMethods = filter(lambda each: callable(getattr(self, each, None)), allMyAttributes)
        myGtViews = filter(lambda each: getattr(getattr(self, each, None), "gtView", False), allMyMethods)
        # those defined by the object that I wrap
        allObjectAttributes = dir(self.object)
        allObjectMethods = list(filter(lambda each: callable(getattr(self.object, each, None)), allObjectAttributes))
        objectGtViews = filter(lambda each: getattr(getattr(self.object, each, None), "gtView", False),
                               allObjectMethods)
        # for now, we keep the old prefix matching, avoiding doubles
        objectGtViews = set(objectGtViews)
        objectGtViews.update(filter(lambda each: each.startswith("gtView"), allObjectMethods))
        # combined into a list of strings
        result = list(set(list(myGtViews) + list(objectGtViews)))
        # when I wrap a @gtView marked method I get false positives,
        # this is normal but they are not real gtViews, remove them
        # my attribute 'object' matches because it is callable and marked
        if 'object' in result:
            result.remove('object')
        # the object's attribute '__func__' matches because it is callable and marked
        if '__func__' in result:
            result.remove('__func__')
        return result

    def getView(self, viewName):
        myView = getattr(self, viewName, None)
        objectView = getattr(self.object, viewName, None)
        if objectView is not None:
            return objectView(ViewBuilder())
        if myView is not None:
            return myView(ViewBuilder())
        raise Exception(f"Cannot find view {viewName}")

    def getDataSource(self, viewName):
        return self.getView(viewName).dataSource()

    def getViewDeclaration(self, viewName):
        try:
            view = self.getView(viewName)
            exportData = view.asDictionaryForExport()
        except Exception as err:
            view = ErrorView()
            view.set_errorMessage(str(err))
            exportData = view.asDictionaryForExport()
        exportData["methodSelector"] = viewName
        return exportData

    def getViewsDeclarations(self):
        viewNames = self.getGtViewMethodNames()
        viewDeclarations = map(lambda each: self.getViewDeclaration(each), viewNames)
        nonEmptyViewDeclarations = filter(lambda each: each["viewName"] != "empty", viewDeclarations)
        return list(nonEmptyViewDeclarations)

    def attributesFor(self, anObject, callables):
        allAttributes = dir(anObject)
        filteredAttributes = filter(lambda each: callable(getattr(anObject, each, None)) == callables, allAttributes)
        keyValuePairs = map(lambda each: [each, getattr(anObject, each, "")], filteredAttributes)
        return list(keyValuePairs)


    @gtView
    def gtViewAttributes(self, aBuilder):
        return aBuilder.columnedList() \
            .title("Raw") \
            .priority(200) \
            .items(lambda: self.attributesFor(self.object, False)) \
            .column("Name", lambda each: each[0]) \
            .column("Value", lambda each: each[1]) \
            .set_accessor(lambda selection: self.attributesFor(self.object, False)[selection][1])

    @gtView
    def gtViewMethods(self, aBuilder):
        return aBuilder.columnedList() \
            .title("Methods") \
            .priority(250) \
            .items(lambda: self.attributesFor(self.object, True)) \
            .column("Name", lambda each: each[0]) \
            .column("Value", lambda each: each[1]) \
            .set_accessor(lambda selection: self.attributesFor(self.object, True)[selection][1])

    @gtView
    def gtViewPrint(self, aBuilder):
        return aBuilder.textEditor() \
            .title("Print") \
            .priority(300) \
            .setString(str(self.object))

    @gtView
    def gtViewList(self, aBuilder):
        if type(self.object) is not list:
            return aBuilder.empty()
        return aBuilder.list()\
            .title("Items")\
            .priority(150)\
            .items(lambda: self.object)\
            .itemFormat(lambda each: str(each))

    @gtView
    def gtViewTuple(self, aBuilder):
        if type(self.object) is not tuple:
            return aBuilder.empty()
        return aBuilder.list()\
            .title("Items")\
            .priority(150)\
            .items(lambda: self.object)\
            .itemFormat(lambda each: str(each))

    @gtView
    def gtViewRange(self, aBuilder):
        if type(self.object) is not range:
            return aBuilder.empty()
        return aBuilder.list()\
            .title("Range")\
            .priority(150)\
            .items(lambda: list(self.object))\
            .itemFormat(lambda each: str(each))

    @gtView
    def gtViewSet(self, aBuilder):
        if type(self.object) is not set:
            return aBuilder.empty()
        return aBuilder.list()\
            .title("Set")\
            .priority(150)\
            .items(lambda: list(self.object))\
            .itemFormat(lambda each: str(each))

    @gtView
    def gtViewInteger(self, aBuilder):
        if type(self.object) is not int:
            return aBuilder.empty()
        return aBuilder.textEditor()\
            .title("Integer")\
            .priority(150)\
            .setString(str(self.object))
 
    @gtView
    def gtViewFloat(self, aBuilder):
        if type(self.object) is not float:
            return aBuilder.empty()
        return aBuilder.textEditor()\
            .title("Float")\
            .priority(150)\
            .setString(str(self.object))

    @gtView
    def gtViewString(self, aBuilder):
        if type(self.object) is not str:
            return aBuilder.empty()
        return aBuilder.textEditor()\
            .title("String")\
            .priority(150)\
            .setString(str(self.object))

    @gtView
    def gtViewBoolean(self, aBuilder):
        if type(self.object) is not bool:
            return aBuilder.empty()
        return aBuilder.textEditor()\
            .title("Boolean")\
            .priority(150)\
            .setString(str(self.object))

    @gtView
    def gtViewNone(self, aBuilder):
        if self.object is not None:
            return aBuilder.empty()
        return aBuilder.textEditor()\
            .title("None")\
            .priority(150)\
            .setString("None")

    @gtView
    def gtViewDict(self, aBuilder):
        if type(self.object) is not dict:
            return aBuilder.empty()
        return aBuilder.columnedList()\
            .title("Items")\
            .priority(150)\
            .items(lambda: self.object.items())\
            .column("Key", lambda each: each[0])\
            .column("Value", lambda each: each[1])
